#!/usr/bin/perl

# Copyright (C) 2002 Bartosz M. Krajnik <bmk@bmk.bz>
#
# This program is free software; you can redistribute it and/or modify     #
# it under the terms of the GNU General Public License as published by     #
# the Free Software Foundation; either version 2 of the License, or        #
# (at your option) any later version.                                      #
#                                                                          #
# This program is distributed in the hope that it will be useful,          #
# but WITHOUT ANY WARRANTY; without even the implied warranty of           #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
# GNU General Public License for more details.                             #
#                                                                          #
# You should have received a copy of the GNU General Public License        #
# along with this program; if not, write to the Free Software              #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA #
# You can also view it online at http://www.gnu.org/licenses/gpl.html      #


##########################################################################
#	Conf section
##########################################################################

#file where syslogd logs data for authfail (see /etc/syslog.conf)
$logfile="/dev/authfail";

#failed logins database
$failfile="/var/log/authfail";

#how many times ONE host can try to fail login
$failled=4;

#don't add these IP numbers into database (You CAN'T describe subnet address)
#don't remember to add there Your IP's (I didn't write time dependency - especially)
#good cracker can wait many hours :)
#$dont_add_into_database = "127.0.0.1, 192.168.0.1, 192.168.0.2";
$dont_add_into_database = "127.0.0.1";

#at the end of this value func update_iptables adds IP
#$command_iptables="/sbin/iptables -I INPUT -p tcp --dport 22 -j DROP -s ";
$command_iptables="/sbin/iptables -I INPUT -j DROP -s ";

#facility for syslog (man syslog.conf)
$facility="daemon";
##########################################################################

use POSIX 'setsid';
use Fcntl;
use Sys::Syslog;

check_files();
daemon_init();

while(1) {
  read_logfile();
}

sub check_files {
  sysopen(TMP, "$logfile", O_NONBLOCK) or die "Can't open FIFO $logfile file (RTFM): $!\n";
  close(TMP);

  open(TMP, "$failfile") or die "Can't open $failfile file (RTFM): $!\n";
  close(TMP);
}

sub daemon_init {
  chdir '/' or die "Can't change chdir to /: $!";
  open STDIN, '/dev/null' or die "Can't read /dev/null: $!";
  open STDOUT, '>/dev/null' or die "Can't write to /dev/null: $!";
  open STDERR, '>&STDOUT' or die "Can't dup stdout: $!";

  if (fork()) {
    exit(0);
  }

  setsid or die "Can't start a new session: $!";
  local $SIG{HUP} = 'IGNORE';

  #after this daemon can't take control terinal in SVR4
  if (fork()) {
    exit(0);
  }
  
  openlog('authfail', 'pid', "$facility");
}

sub update_iptables{
  $ipd = shift;

  $status = system "$command_iptables $ipd";
  syslog("info", "$command_iptables $ipd");

  if($status != 0) {
    syslog("err", "Can't rebuild Netfilter (RTFM)");
    die();
  }

}

sub read_logfile {
  open(LOGFILE, "$logfile") or syslog("err", "Can't read from file $logfile") and die();

  while ($txt = <LOGFILE>) {
    chomp($txt);

    #Failed password for invalid user (.*) from (.*) port (.*)$
    #Failed password for (.*) from (.*) port (.*)$
    #authentication failed for user (.*) - (.*)$
    #Illegal user (.*) from (.*)$
    if($txt =~ /(Failed password for invalid user|Failed password for) (.*) from (.*) port (.*)$/
        or
       $txt =~ /(authentication failed for user|authentication failed: no such user:) (.*) - (.*)$/
        or
       $txt =~ /(Illegal user) (.*) from (.*)$/
        or
       $txt =~ /(authentication) failure;(.*) rhost=(.*) user=(.*)$/) {
      $ip = $3;
      if(update_failfile($ip)) {
        update_iptables($ip);
      }
    }

  }

}

sub update_failfile {
  $ip_tmp = shift;
  $is_in_tmp = 0;
  $ret_value = 0;

  if($dont_add_into_database =~ /$ip_tmp/) {
    return(0);
  }
  
  open (FAILFILE, "$failfile") or syslog("err", "Can't read from file $failfile") and die();
  while($txt = <FAILFILE>) {
    if($txt =~ /^$ip_tmp\ $failled$/) {
      close(FAILFILE);
      return(0);
    }
  }
  close(FAILFILE);


  open (FAILFILE, "+< $failfile") or syslog("err", "Can't read from file $failfile") and die();
  $out = '';

  while($txt = <FAILFILE>) {

    if($txt =~ /^$ip_tmp\ (.*)$/) {
      $attempts=$1;
      $attempts += 1;
      if($attempts == $failled) {
        $ret_value = 1;
      }
      $out .= sprintf "$ip_tmp $attempts\n";
      syslog("info", "$ip_tmp $attempts - updated attempts");
      $is_in_tmp = 1;
    } else {
      $out .= $txt;
    }

  }

  if(! $is_in_tmp) {
    $out .= sprintf "$ip_tmp 1\n";
    syslog("info", "$ip_tmp 1 - added to $failfile");
  }

  seek(FAILFILE, 0, 0) or syslog("err", "Can't seek() in file $failfile") and die();
  print FAILFILE $out or syslog("err", "Can't print() to file $failfile") and die();
  truncate(FAILFILE, tell(FAILFILE)) or syslog("err", "Can't truncate() file $failfile") and die();
  close(FAILFILE);

  if($is_in_tmp) {
    return($ret_value);
  } 

  return(0);
}
